
#if !defined(_NativeInterpreterFloat__)
#define _NativeInterpreterFloat__

#include "milieu.h"
#include "softfloat.h"
#include "nanojit.h"

namespace nanojit
{

struct float32_t
{
	float32 v;
};

struct float64_t
{
	float64 v;
};

inline float32_t operator-(float32_t lhs) {
	float32_t res = { float32_neg(lhs.v) };
	return res;
}

inline float32_t operator+(float32_t lhs, float32_t rhs) {
	float32_t res = { float32_add(lhs.v, rhs.v) };
	return res;
}

inline float32_t operator-(float32_t lhs, float32_t rhs) {
	float32_t res = { float32_sub(lhs.v, rhs.v) };
	return res;
}

inline float32_t operator*(float32_t lhs, float32_t rhs) {
	float32_t res = { float32_mul(lhs.v, rhs.v) };
	return res;
}

inline float32_t operator/(float32_t lhs, float32_t rhs) {
	float32_t res = { float32_div(lhs.v, rhs.v) };
	return res;
}

inline bool operator==(float32_t lhs, float32_t rhs) {
	flag res = float32_eq(lhs.v, rhs.v);
	return (res != 0);
}

inline bool operator!=(float32_t lhs, float32_t rhs) {
	flag res = float32_eq(lhs.v, rhs.v);
	return (res == 0);
}

inline bool operator<(float32_t lhs, float32_t rhs) {
	flag res = float32_lt(lhs.v, rhs.v);
	return (res != 0);
}

inline bool operator<=(float32_t lhs, float32_t rhs) {
	flag res = float32_le(lhs.v, rhs.v);
	return (res != 0);
}

inline bool operator>(float32_t lhs, float32_t rhs) {
	flag res = float32_lt(rhs.v, lhs.v);
	return (res != 0);
}

inline bool operator>=(float32_t lhs, float32_t rhs) {
	flag res = float32_le(rhs.v, lhs.v);
	return (res != 0);
}

inline float32_t abs(float32_t lhs) {
	float32_t res = { float32_abs(lhs.v) };
	return res;
}

inline float32_t sqrt(float32_t v) {
	float32_t res = { float32_sqrt(v.v) };
	return res;
}

inline float32_t rcp(float32_t v) {
	float32_t res = { float32_div(int32_to_float32(1), v.v) };
	return res;
}

inline float32_t rsqrt(float32_t v) {
	float32_t res = { float32_div(int32_to_float32(1), float32_sqrt(v.v)) };
	return res;
}

inline float32_t mulneg(float32_t a, float32_t b) {
	float32_t r;
	r.v = float32_mulneg(a.v, b.v);
	return r;
}



// float64 operators

inline float64_t operator-(float64_t lhs) {
	float64_t res = { float64_neg(lhs.v) };
	return res;
}

inline float64_t operator+(float64_t lhs, float64_t rhs) {
	float64_t res = { float64_add(lhs.v, rhs.v) };
	return res;
}

inline float64_t operator-(float64_t lhs, float64_t rhs) {
	float64_t res = { float64_sub(lhs.v, rhs.v) };
	return res;
}

inline float64_t operator*(float64_t lhs, float64_t rhs) {
	float64_t res = { float64_mul(lhs.v, rhs.v) };
	return res;
}

inline float64_t operator/(float64_t lhs, float64_t rhs) {
	float64_t res = { float64_div(lhs.v, rhs.v) };
	return res;
}

inline bool operator==(float64_t lhs, float64_t rhs) {
	flag res = float64_eq(lhs.v, rhs.v);
	return (res != 0);
}

inline bool operator!=(float64_t lhs, float64_t rhs) {
	flag res = float64_eq(lhs.v, rhs.v);
	return (res == 0);
}

inline bool operator<(float64_t lhs, float64_t rhs) {
	flag res = float64_lt(lhs.v, rhs.v);
	return (res != 0);
}

inline bool operator<=(float64_t lhs, float64_t rhs) {
	flag res = float64_le(lhs.v, rhs.v);
	return (res != 0);
}

inline bool operator>(float64_t lhs, float64_t rhs) {
	flag res = float64_lt(rhs.v, lhs.v);
	return (res != 0);
}

inline bool operator>=(float64_t lhs, float64_t rhs) {
	flag res = float64_le(rhs.v, lhs.v);
	return (res != 0);
}

inline float64_t abs(float64_t lhs) {
	float64_t res = { float64_abs(lhs.v) };
	return res;
}

inline float64_t sqrt(float64_t v) {
	float64_t res = { float64_sqrt(v.v) };
	return res;
}

inline float64_t rcp(float64_t v) {
	float64_t res = { float64_div(int32_to_float64(1), v.v) };
	return res;
}

inline float64_t rsqrt(float64_t v) {
	float64_t res = { float64_div(int32_to_float64(1), float64_sqrt(v.v)) };
	return res;
}

inline float64_t mulneg(float64_t a, float64_t b) {
	float64_t r;
	r.v = float64_mulneg(a.v, b.v);
	return r;
}

// 

template<typename TS, typename TD, bool a_bRoundToZero>
TD convert(TS v);

// float32 conversions

template<>
inline int32_t convert<float32_t, int32_t, false>(float32_t v) {
	return float32_to_int32(v.v);
}

template<>
inline int64_t convert<float32_t, int64_t, false>(float32_t v) {
	return float32_to_int64(v.v);
}

template<>
inline uint32_t convert<float32_t, uint32_t, false>(float32_t v) {
	int64_t res = float32_to_int64(v.v);
	if (res < 0ULL)
		return 0UL;
	else if (res > 0xFFFFFFFFULL)
		return 0xFFFFFFFFUL;
	return static_cast<uint32_t>(res);
}

template<>
inline uint64_t convert<float32_t, uint64_t, false>(float32_t v) {
    NanoAssert(0);
	return float32_to_int64(v.v);
}

template<>
inline int32_t convert<float32_t, int32_t, true>(float32_t v) {
	return float32_to_int32_round_to_zero(v.v);
}

template<>
inline int64_t convert<float32_t, int64_t, true>(float32_t v) {
	return float32_to_int64_round_to_zero(v.v);
}

template<>
inline uint32_t convert<float32_t, uint32_t, true>(float32_t v) {
	int64_t res = float32_to_int64_round_to_zero(v.v);
	if (res < 0ULL)
		return 0UL;
	else if (res > 0xFFFFFFFFULL)
		return 0xFFFFFFFFUL;
	return static_cast<uint32_t>(res);
}

template<>
inline uint64_t convert<float32_t, uint64_t, true>(float32_t v) {
	NanoAssert(0);
	return float32_to_int64_round_to_zero(v.v);
}

template<>
inline float64_t convert<float32_t, float64_t, false>(float32_t v) {
	float64_t res = { float32_to_float64(v.v) };
	return res;
}

template<>
inline float32_t convert<int32_t, float32_t, false>(int32_t v) {
	float32_t res = { int32_to_float32(v) };
	return res;
}

template<>
inline float32_t convert<uint32_t, float32_t, false>(uint32_t v) {
	int64_t tmp = static_cast<int64_t>(v);
	float32_t res = { int64_to_float32(tmp) };
	return res;
}


template<>
inline float32_t convert<int64_t, float32_t, false>(int64_t v) {
	float32_t res = { int64_to_float32(v) };
	return res;
}

template<>
inline float32_t convert<uint64_t, float32_t, false>(uint64_t v) {
	NanoAssert(0);
	float32_t res = { int64_to_float32(v) };
	return res;
}

// float64 conversions

template<>
inline int32_t convert<float64_t, int32_t, false>(float64_t v) {
	return float64_to_int32(v.v);
}

template<>
inline int64_t convert<float64_t, int64_t, false>(float64_t v) {
	return float64_to_int64(v.v);
}

template<>
inline uint32_t convert<float64_t, uint32_t, false>(float64_t v) {
	int64_t res = float64_to_int64(v.v);
	if (res < 0ULL)
		return 0UL;
	else if (res > 0xFFFFFFFFULL)
		return 0xFFFFFFFFUL;
	return static_cast<uint32_t>(res);
}

template<>
inline uint64_t convert<float64_t, uint64_t, false>(float64_t v) {
	NanoAssert(0);
	return float64_to_int64(v.v);
}



template<>
inline int32_t convert<float64_t, int32_t, true>(float64_t v) {
	return float64_to_int32_round_to_zero(v.v);
}

template<>
inline int64_t convert<float64_t, int64_t, true>(float64_t v) {
	return float64_to_int64_round_to_zero(v.v);
}

template<>
inline uint32_t convert<float64_t, uint32_t, true>(float64_t v) {
	int64_t res = float64_to_int64_round_to_zero(v.v);
	if (res < 0ULL)
		return 0UL;
	else if (res > 0xFFFFFFFFULL)
		return 0xFFFFFFFFUL;
	return static_cast<uint32_t>(res);
}

template<>
inline uint64_t convert<float64_t, uint64_t, true>(float64_t v) {
	NanoAssert(0);
	return float64_to_int64_round_to_zero(v.v);
}

template<>
inline float32_t convert<float64_t, float32_t, false>(float64_t v) {
	float32_t res = { float64_to_float32(v.v) };
	return res;
}

template<>
inline float64_t convert<int32_t, float64_t, false>(int32_t v) {
	float64_t res = { int32_to_float64(v) };
	return res;
}

template<>
inline float64_t convert<uint32_t, float64_t, false>(uint32_t v) {
	int64_t tmp = static_cast<int64_t>(v);
	float64_t res = { int64_to_float64(tmp) };
	return res;
}

template<>
inline float64_t convert<int64_t, float64_t, false>(int64_t v) {
	float64_t res = { int64_to_float64(v) };
	return res;
}

template<>
inline float64_t convert<uint64_t, float64_t, false>(uint64_t v) {
	NanoAssert(0);
	float64_t res = { int64_to_float64(v) };
	return res;
}





} // namespace nanojit

#endif // !defined(_NativeInterpreterFloat__)

