

#include "nanojit.h"

namespace nanojit
{

#define INVOLUTION(ins, oprnd, suffix)\
	if (ins->opcode() == oprnd->opcode())\
		return ins->initLInsOp1(LIR_copy##suffix, oprnd->oprnd1());



void ExprLirOptimizationPass::ins1(LIns* ins)
{
	LOpcode     v       = ins->opcode();
	LIns*       oprnd   = ins->oprnd1();

	if (!oprnd->canOptimize())
		return;

	switch (v) {
#ifdef NANOJIT_64BIT
	case LIR_movq2i:
		if (oprnd->isImmQ())
			return ins->initLInsI(static_cast<int32_t>(oprnd->immQ()));
		break;
	case LIR_movsi2sq:
		if (oprnd->isImmI())
			return ins->initLInsQ(int64_t(int32_t(oprnd->immI())));
		break;
	case LIR_movui2uq:
		if (oprnd->isImmI())
			return ins->initLInsQ(int64_t(uint32_t(oprnd->immI())));
		break;
	case LIR_notq:
		if (oprnd->isImmQ())
			return ins->initLInsQ(~(int64_t)oprnd->immQ());
		INVOLUTION(ins, oprnd, q);
		break;
	case LIR_negq:
		if (oprnd->isImmQ())
			return ins->initLInsQ(-(int64_t)oprnd->immQ());
		INVOLUTION(ins, oprnd, q);
		break;
#endif
	case LIR_noti:
		if (oprnd->isImmI())
			return ins->initLInsI(~oprnd->immI());
		INVOLUTION(ins, oprnd, i);
		break;
	case LIR_negi:
		if (oprnd->isImmI())
			return ins->initLInsI(-oprnd->immI());
		if (oprnd->isop(LIR_subi)) // -(a-b) = b-a
			return ins->initLInsOp2(LIR_subi, oprnd->oprnd2(), oprnd->oprnd1());
		INVOLUTION(ins, oprnd, i);
		break;
	case LIR_negd:
		if (oprnd->isImmD())
			return ins->initLInsD(-oprnd->immD());
		if (oprnd->isop(LIR_subd))
			return ins->initLInsOp2(LIR_subd, oprnd->oprnd2(), oprnd->oprnd1());
		INVOLUTION(ins, oprnd, d);
		break;
	case LIR_negf:
		if (oprnd->isImmF())
			return ins->initLInsF(-oprnd->immF());
		if (oprnd->isop(LIR_subf))
			return ins->initLInsOp2(LIR_subf, oprnd->oprnd2(), oprnd->oprnd1());
		INVOLUTION(ins, oprnd, f);
		break;
#if defined NANOJIT_64BIT
	case LIR_movq2d:
		if (oprnd->isImmQ())
			return ins->initLInsD(double(oprnd->immQ()));
		break;
	case LIR_movd2q:
		if (oprnd->isImmD())
			return ins->initLInsQ(int64_t(oprnd->immD()));
		break;
#endif
	case LIR_absd:
		INVOLUTION(ins, oprnd, d);
		break;
	case LIR_absf:
		INVOLUTION(ins, oprnd, f);
		break;

	case LIR_copyi:
		if (oprnd->isImmI())
		{
			return ins->initLInsI(oprnd->immI());
		}
		break;

#if defined NANOJIT_64BIT
	case LIR_copyq:
		if (oprnd->isImmQ())
		{
			return ins->initLInsQ((int64_t)oprnd->immQ());
		}
		break;
#endif

	case LIR_copyf:
		if (oprnd->isImmF())
		{
			return ins->initLInsF(oprnd->immF());
		}
		break;

	case LIR_copyd:
		if (oprnd->isImmD())
		{
			return ins->initLInsD(oprnd->immD());
		}
		break;


	case LIR_getflgz:
		if (oprnd->isImmI())
		{
			return ins->initLInsI(oprnd->immI() == 0 ? 1 : 0);
		}
#ifdef NANOJIT_64BIT
		else if (oprnd->isImmQ())
		{
			return ins->initLInsI(oprnd->immQ() == 0 ? 1 : 0);
		}
#endif
		break;

	case LIR_getflgs:
		if (oprnd->isImmI())
		{
			return ins->initLInsI(static_cast<uint32_t>(oprnd->immI() >> 31));
		}
#ifdef NANOJIT_64BIT
		else if (oprnd->isImmQ())
		{
			return ins->initLInsI(static_cast<uint32_t>(oprnd->immQ() >> 63));
		}
#endif
		break;

	case LIR_getflgc:
	case LIR_getflgb:
#ifdef NANOJIT_64BIT
		if (oprnd->isImmI() || oprnd->isImmQ())
#else
		if (oprnd->isImmI())
#endif
		{
			return ins->initLInsI(0);
		}
		break;

	case LIR_getflgv:
#ifdef NANOJIT_64BIT
		if (oprnd->isImmI() || oprnd->isImmQ())
#else
		if (oprnd->isImmI())
#endif
		{
			return ins->initLInsI(0);
		}
		break;

	case LIR_getflgnz:
		if (oprnd->isImmI())
		{
			return ins->initLInsI(oprnd->immI() != 0 ? 1 : 0);
		}
#ifdef NANOJIT_64BIT
		else if (oprnd->isImmQ())
		{
			return ins->initLInsI(oprnd->immQ() != 0 ? 1 : 0);
		}
#endif
		break;

	case LIR_getflgns:
		if (oprnd->isImmI())
		{
			return ins->initLInsI(static_cast<uint32_t>(1 - (oprnd->immI() >> 31)));
		}
#ifdef NANOJIT_64BIT
		else if (oprnd->isImmQ())
		{
			return ins->initLInsI(static_cast<uint32_t>(1 - (oprnd->immQ() >> 63)));
		}
#endif
		break;

	case LIR_getflgnc:
#ifdef NANOJIT_64BIT
		if (oprnd->isImmI() || oprnd->isImmQ())
#else
		if (oprnd->isImmI())
#endif
		{
			return ins->initLInsI(1);
		}
		break;

	case LIR_getflgnv:
#ifdef NANOJIT_64BIT
		if (oprnd->isImmI() || oprnd->isImmQ())
#else
		if (oprnd->isImmI())
#endif
		{
			return ins->initLInsI(1);
		}
		break;

	default:
		;
	}
}

// This is an ugly workaround for an apparent compiler
// bug; in VC2008, compiling with optimization on
// will produce spurious errors if this code is inlined
// into ExprLirOptimizationPass::ins2(). See https://bugzilla.mozilla.org/show_bug.cgi?id=538504
inline double do_join(int32_t c1, int32_t c2) {
	union {
		double d;
		uint64_t u64;
	} u;
	u.u64 = uint32_t(c1) | uint64_t(c2)<<32;
	return u.d;
}

void ExprLirOptimizationPass::ins2(LIns* ins) {
	LOpcode     v       = ins->opcode();
	LIns*       oprnd1  = ins->oprnd1();
	LIns*       oprnd2  = ins->oprnd2();


	//-------------------------------------------------------------------
	// Folding where the two operands are equal
	//-------------------------------------------------------------------
	if (oprnd1 == oprnd2) {
		// The operands are equal.
		switch (v) {
		case LIR_subi:
		CASE64(LIR_subq:)
			if (oprnd1->canOptimize() && (!ins->getGenFlags()))
				return ins->initLInsI(0);
			break;

		case LIR_xori:
		CASE64(LIR_xorq:)
		case LIR_ltui:
		CASE64(LIR_ltuq:)
		case LIR_gtui:
		CASE64(LIR_gtuq:)
		case LIR_gti:
		CASE64(LIR_gtq:)
		case LIR_lti:
		CASE64(LIR_ltq:)
		case LIR_andci:
		CASE64(LIR_andcq:)
			if (oprnd1->canOptimize())
				return ins->initLInsI(0);
			break;

		case LIR_ori:
		CASE64(LIR_orq:)
		case LIR_andi:
		CASE64(LIR_andq:)
			return ins->initLInsOp1(LIR_copyi, oprnd1);

		case LIR_eqi:
		CASE64(LIR_eqq:)
		case LIR_lei:
		CASE64(LIR_leq:)
		case LIR_leui:
		CASE64(LIR_leuq:)
		case LIR_gei:
		CASE64(LIR_geq:)
		case LIR_geui:
		CASE64(LIR_geuq:)
			// (x <= x) == 1; (x >= x) == 1
			if (oprnd1->canOptimize())
				return ins->initLInsI(1);
			break;

		case LIR_nei:
		CASE64(LIR_neq:)
			if (oprnd1->canOptimize())
				return ins->initLInsI(0);
			break;

		default:
			break;
		}
	}

	//-------------------------------------------------------------------
	// Folding where both operands are immediates, grouped by type
	//-------------------------------------------------------------------
	if (oprnd1->canOptimize() && oprnd1->isImmAny() && 
		oprnd2->canOptimize() && oprnd2->isImmAny() &&
		!ins->getGenFlags()) {

		if (oprnd1->isImmI() && oprnd2->isImmI()) {

			// The operands are both int immediates.
			int32_t c1 = oprnd1->immI();
			int32_t c2 = oprnd2->immI();

			switch (v) {
			case LIR_eqi:   return ins->initLInsI(c1 == c2);
			case LIR_nei:   return ins->initLInsI(c1 != c2);
			case LIR_lti:   return ins->initLInsI(c1 <  c2);
			case LIR_gti:   return ins->initLInsI(c1 >  c2);
			case LIR_lei:   return ins->initLInsI(c1 <= c2);
			case LIR_gei:   return ins->initLInsI(c1 >= c2);
			case LIR_ltui:  return ins->initLInsI(uint32_t(c1) <  uint32_t(c2));
			case LIR_gtui:  return ins->initLInsI(uint32_t(c1) >  uint32_t(c2));
			case LIR_leui:  return ins->initLInsI(uint32_t(c1) <= uint32_t(c2));
			case LIR_geui:  return ins->initLInsI(uint32_t(c1) >= uint32_t(c2));

			case LIR_lshi:  return ins->initLInsI(c1 << (c2 & 0x1f));
			case LIR_rshi:  return ins->initLInsI(c1 >> (c2 & 0x1f));
			case LIR_rshui: return ins->initLInsI(int32_t(uint32_t(c1) >> (c2 & 0x1f)));

			case LIR_ori:   return ins->initLInsI(c1 | c2);
			case LIR_andi:  return ins->initLInsI(c1 & c2);
			case LIR_andci: return ins->initLInsI(c1 & ~c2);
			case LIR_xori:  return ins->initLInsI(c1 ^ c2);

			case LIR_addi:  return ins->initLInsI(c1 + c2);
			case LIR_subi:  return ins->initLInsI(c1 - c2);
			case LIR_muli:  return ins->initLInsI(c1 * c2);

	#if defined(NANOJIT_HOST_PLATFORM_X64)
			case LIR_diviu:
			case LIR_divis:
			case LIR_modi:
				// We can't easily fold div and mod, since folding div makes it
				// impossible to calculate the mod that refers to it. The
				// frontend shouldn't emit div and mod with constant operands.
				NanoAssert(0);
	#endif

			default:
				break;
			}

#ifdef NANOJIT_64BIT
		} else if ((oprnd1->isImmQ() || oprnd1->isImmI()) && (oprnd2->isImmQ() || oprnd2->isImmI())) {

			// The operands are both quad immediates.
			int64_t c1 = oprnd1->isImmQ() ? oprnd1->immQ() : oprnd1->immI();
			int64_t c2 = oprnd2->isImmQ() ? oprnd2->immQ() : oprnd2->immI();

			switch (v) {
			case LIR_eqq:   return ins->initLInsI(c1 == c2);
			case LIR_neq:   return ins->initLInsI(c1 != c2);
			case LIR_ltq:   return ins->initLInsI(c1 <  c2);
			case LIR_gtq:   return ins->initLInsI(c1 >  c2);
			case LIR_leq:   return ins->initLInsI(c1 <= c2);
			case LIR_geq:   return ins->initLInsI(c1 >= c2);
			case LIR_ltuq:  return ins->initLInsI(uint64_t(c1) <  uint64_t(c2));
			case LIR_gtuq:  return ins->initLInsI(uint64_t(c1) >  uint64_t(c2));
			case LIR_leuq:  return ins->initLInsI(uint64_t(c1) <= uint64_t(c2));
			case LIR_geuq:  return ins->initLInsI(uint64_t(c1) >= uint64_t(c2));

			case LIR_orq:   return ins->initLInsQ(c1 | c2);
			case LIR_andq:  return ins->initLInsQ(c1 & c2);
			case LIR_andcq: return ins->initLInsQ(c1 & ~c2);
			case LIR_xorq:  return ins->initLInsQ(c1 ^ c2);

			case LIR_addq:  return ins->initLInsQ(c1 + c2);
			case LIR_subq:  return ins->initLInsQ(c1 - c2);
			case LIR_mulq:  return ins->initLInsQ(c1 * c2);

			case LIR_lshq:  return ins->initLInsQ(c1 << (c2 & 0x3f));
			case LIR_rshq:  return ins->initLInsQ(c1 >> (c2 & 0x3f));
			case LIR_rshuq: return ins->initLInsQ(int64_t(uint64_t(c1) >> (c2 & 0x3f)));

			default:
				break;
			}
#endif  // NANOJIT_64BIT

		} else if (oprnd1->isImmD() && oprnd2->isImmD()) {
			// The operands are both double immediates.
			double c1 = oprnd1->immD();
			double c2 = oprnd2->immD();
			switch (v) {
			case LIR_eqd:   return ins->initLInsI(c1 == c2);
			case LIR_ned:   return ins->initLInsI(c1 != c2);
			case LIR_ltd:   return ins->initLInsI(c1 <  c2);
			case LIR_gtd:   return ins->initLInsI(c1 >  c2);
			case LIR_led:   return ins->initLInsI(c1 <= c2);
			case LIR_ged:   return ins->initLInsI(c1 >= c2);

			case LIR_addd:  return ins->initLInsD(c1 + c2);
			case LIR_subd:  return ins->initLInsD(c1 - c2);
			case LIR_muld:  return ins->initLInsD(c1 * c2);
			case LIR_divd:  return ins->initLInsD(c1 / c2);

			default:        break;
			}
		} else if (oprnd1->isImmF() && oprnd2->isImmF()) {
			// The operands are both single-precision float immediates.
			float c1 = oprnd1->immF();
			float c2 = oprnd2->immF();
			switch (v) {
			case LIR_eqf:   return ins->initLInsI(c1 == c2);
			case LIR_nef:   return ins->initLInsI(c1 != c2);
			case LIR_ltf:   return ins->initLInsI(c1 <  c2);
			case LIR_gtf:   return ins->initLInsI(c1 >  c2);
			case LIR_lef:   return ins->initLInsI(c1 <= c2);
			case LIR_gef:   return ins->initLInsI(c1 >= c2);

			case LIR_addf:  return ins->initLInsF(c1 + c2);
			case LIR_subf:  return ins->initLInsF(c1 - c2);
			case LIR_mulf:  return ins->initLInsF(c1 * c2);
			case LIR_divf:  return ins->initLInsF(c1 / c2);

			default:        break;
			}
		}
	}

	//-------------------------------------------------------------------
	// If only one operand is an immediate, make sure it's on the RHS, if possible
	//-------------------------------------------------------------------
	if (oprnd1->isImmAny() && !oprnd2->isImmAny()) {
		switch (v) {
		case LIR_eqi:
		CASE64(LIR_eqq:)
		case LIR_eqd:
		case LIR_eqf:
		case LIR_nei:
		CASE64(LIR_neq:)
		case LIR_nef:
		case LIR_ned:
		case LIR_addi:
		CASE64(LIR_addq:)
		case LIR_addd:
		case LIR_addf:
		case LIR_muli:
		CASE64(LIR_mulq:)
		case LIR_muld:
		case LIR_mulf:
		case LIR_andi:
		CASE64(LIR_andq:)
		case LIR_ori:
		CASE64(LIR_orq:)
		case LIR_xori:
		CASE64(LIR_xorq:)
		case LIR_mullui:
		CASE64(LIR_mulluq:)
		case LIR_mullsi:
		CASE64(LIR_mullsq:)
		{
			// move immediate to RHS
			LIns* t = oprnd2;
			oprnd2 = oprnd1;
			oprnd1 = t;
			ins->initLInsOp2(v, oprnd1, oprnd2);
			break;
		}

		case LIR_lti:
		CASE64(LIR_ltq:)
		case LIR_ltf:
		case LIR_ltd:
		case LIR_gei:
		CASE64(LIR_geq:)
		case LIR_gef:
		case LIR_ged:
		case LIR_gti:
		CASE64(LIR_gtq:)
		case LIR_gtf:
		case LIR_gtd:
		case LIR_lei:
		CASE64(LIR_leq:)
		case LIR_lef:
		case LIR_ltui:
		case LIR_led:
		CASE64(LIR_ltuq:)
		case LIR_geui:
		CASE64(LIR_geuq:)
		case LIR_gtui:
		CASE64(LIR_gtuq:)
		case LIR_leui:
		CASE64(LIR_leuq:)
		{
			LIns* t = oprnd2;
			oprnd2 = oprnd1;
			oprnd1 = t;
			v = swapCmpOpcode(v);
			ins->initLInsOp2(v, oprnd1, oprnd2);
			break;
		}

		default:
			break;
		}
	}

	//-------------------------------------------------------------------
	// Folding where the RHS is an immediate
	//-------------------------------------------------------------------
	if (oprnd2->isImmAny() && oprnd2->canOptimize() && (!ins->getGenFlags())) {

		if (oprnd2->isImmI()) {
			// The second operand is an int immediate.
			int c = oprnd2->immI();

			if (c == 0) {
				switch (v) {
				case LIR_addi:
				case LIR_ori:
				case LIR_xori:
				case LIR_subi:
				case LIR_lshi:
				case LIR_rshi:
				case LIR_rshui:
				case LIR_rori:
				case LIR_rcri:
					return ins->initLInsOp1(LIR_copyi, oprnd1);

	#if defined(NANOJIT_64BIT)
				CASE64(LIR_lshq:)   // These are here because their RHS is an int
				CASE64(LIR_rshq:)
				CASE64(LIR_rshuq:)
				CASE64(LIR_rorq:)
				CASE64(LIR_rcrq:)
					return ins->initLInsOp1(LIR_copyq, oprnd1);
	#endif

				case LIR_andi:
				case LIR_muli:
				case LIR_ltui: // unsigned < 0 -> always false
					return ins->initLInsI(0);

				case LIR_andci:
					return ins->initLInsOp1(LIR_copyi, oprnd1);

				case LIR_geui: // unsigned >= 0 -> always true
					return ins->initLInsI(1);

				default:
					break;
				}

			} else if (c == -1) {
				switch (v) {
				case LIR_ori:   return ins->initLInsI(c); //*ins = *oprnd2; return;       // x | -1 = -1
				case LIR_andi:  return ins->initLInsOp1(LIR_copyi, oprnd1); //*ins = *oprnd1; return;       // x & -1 = x
				case LIR_andci: return ins->initLInsI(0);   // u32 & ~0xffffffff -> always 0
				case LIR_gtui:  return ins->initLInsI(0);   // u32 >  0xffffffff -> always false
				case LIR_leui:  return ins->initLInsI(1);   // u32 <= 0xffffffff -> always true
				default:        break;
				}

			} else if (c == 1) {
				switch (v) {
					case LIR_muli:  return ins->initLInsOp1(LIR_copyi, oprnd1);
					default:        break;
				}
			}

#ifdef NANOJIT_64BIT
		} else if (oprnd2->isImmQ()) {
			// The second operand is a quad immediate.
			int64_t c = oprnd2->immQ();
			if (c == 0) {
				switch (v) {
				case LIR_addq:
				case LIR_orq:
				case LIR_xorq:
				case LIR_subq:
					return ins->initLInsOp1(LIR_copyq, oprnd1);

				case LIR_andq:
					return ins->initLInsI(0);

				case LIR_andcq: // always 0
					return ins->initLInsOp1(LIR_copyq, oprnd2);

				case LIR_ltuq: // unsigned < 0 -> always false
					return ins->initLInsI(0);

				case LIR_geuq: // unsigned >= 0 -> always true
					return ins->initLInsI(1);

				default:
					break;
				}

			} else if (c == -1) {
				switch (v) {
				case LIR_orq:   return ins->initLInsOp1(LIR_copyq, oprnd2); //*ins = *oprnd2; return;       // x | -1 = -1
				case LIR_andq:  return ins->initLInsOp1(LIR_copyq, oprnd1); //*ins = *oprnd1; return;       // x & -1 = x
				case LIR_andcq: return ins->initLInsI(0);   // always 0
				case LIR_gtuq:  return ins->initLInsI(0);   // u64 >  0xffffffffffffffff -> always false
				case LIR_leuq:  return ins->initLInsI(1);   // u64 <= 0xffffffffffffffff -> always true
				default:       break;
				}

			} else if (c == 1) {
				if (v == LIR_mulq) {
					return ins->initLInsOp1(LIR_copyq, oprnd1);
				}
			}
#endif  // NANOJIT_64BIT
		}
	}

}

void ExprLirOptimizationPass::ins3(LIns* ins) {
	LOpcode     v       = ins->opcode();
	LIns*       oprnd1  = ins->oprnd1();
	LIns*       oprnd2  = ins->oprnd2();
	LIns*       oprnd3  = ins->oprnd3();

	NanoAssert(oprnd1 && oprnd2 && oprnd3);

	switch (v) {
		CASE64(LIR_cmovq:)
		case LIR_cmovf:
		case LIR_cmovi:
		case LIR_cmovd:
		{

			LOpcode     copyOpcode  = LIR_sentinel;
			switch (v)
			{
			case LIR_cmovi:     copyOpcode = LIR_copyi;     break;
#if defined NANOJIT_64BIT
			case LIR_cmovq:     copyOpcode = LIR_copyq;     break;
#endif
			case LIR_cmovf:     copyOpcode = LIR_copyf;     break;
			case LIR_cmovd:     copyOpcode = LIR_copyd;     break;
			default:            NanoAssert(0);              break;
			}
			if (oprnd2 == oprnd3) {
				// c ? a : a => a
				return ins->initLInsOp1(copyOpcode, oprnd2); // *ins = *oprnd2;     return;
			}
			if (oprnd1->isImmI() && oprnd1->canOptimize()) {
				// immediate ? x : y => return x or y depending on immediate
				return ins->initLInsOp1(copyOpcode, oprnd1->immI() ? oprnd2 : oprnd3); //ins = oprnd1->immI() ? oprnd2 : oprnd3; return;
			}
			break;
		}

		default:
			NanoAssert(0); // invalid opcode
			break;
	}
}

void ExprLirOptimizationPass::insBranch(LIns *ins) {
	LOpcode v = ins->opcode();
	LIns* c = ins->branchCondition();
	LIns* t = ins->branchTarget();

	if ((c == NULL) || (!c->canOptimize()))
		return;

	if (v == LIR_jt || v == LIR_jf) {
		if (c->isImmI()) {
			if ((v == LIR_jt && !c->immI()) || (v == LIR_jf && c->immI())) {
				// no jump needed
				return ins->initLInsOp0(LIR_nop);
			} else {
#ifdef JS_TRACER
				// We're emitting a branch that will always be taken.  This may
				// result in dead code that will not be optimized away, and
				// could indicate a performance problem or other bug, so assert
				// in debug builds.
				NanoAssertMsg(0, "Constantly taken branch detected");
#endif
				ins->initLInsBranch(LIR_j, NULL);
				ins->branchSetTarget(t);
			}
		} else {
			while (c->isop(LIR_eqi) && c->oprnd1()->isCmp() && c->oprnd2()->isImmI(0)) {
				// jt(eq(cmp,0)) => jf(cmp)   or   jf(eq(cmp,0)) => jt(cmp)
				v = invertCondJmpOpcode(v);
				//c->initLInsOp2(LIR_eqi, c->oprnd1(), c->oprnd2()); //*c = *c->oprnd1();
				return ins->initLInsOp2(v, c->oprnd1(), t);
			}
		}
	}
}

bool ExprLirOptimizationPass::insIsS16(LIns* i) {
	if (i->isImmI()) {
		int c = i->immI();
		return isS16(c);
	}
	if (i->isCmov()) {
		return insIsS16(i->oprnd2()) && insIsS16(i->oprnd3());
	}
	if (i->isCmp())
		return true;
	// many other possibilities too.
	return false;
}



void ExprLirOptimizationPass::processLIns(LIns* src) {
	LInsRepKind kind = src->repKind();
	switch (kind) {
		case LRK_Branch:
			NanoAssert(src->isLInsBranch());
			insBranch(src);
			break;

		case LRK_Op1:
			NanoAssert(src->isLInsOp1());
			ins1(src);
			break;

		case LRK_Op2:
			NanoAssert(src->isLInsOp2());
			ins2(src);
			break;

		case LRK_Op3:
			NanoAssert(src->isLInsOp3());
			ins3(src);
			break;

		default:
			NanoAssert(!src->isLInsBranch());
			NanoAssert(!src->isLInsOp1());
			NanoAssert(!src->isLInsOp2());
			NanoAssert(!src->isLInsOp3());
			break;
	}
}

}


