

#include "nanojit.h"


namespace nanojit
{

HeapBlock* HeapBlock::setupHeapBlockFromBuffer(void* buffer, size_t size, HeapBlock* next) {

	size_t heapBlockAlignMask = getTypeAlign<HeapBlock>() - 1;
	size_t heapBlockoffset = (size - sizeof(HeapBlock)) & ~heapBlockAlignMask;
	HeapBlock* heapBlock = reinterpret_cast<HeapBlock*>(reinterpret_cast<uint8_t*>(buffer) + heapBlockoffset);

	heapBlock->initialize(buffer, size, next);

	return heapBlock;
}


void HeapBlock::initialize(void* buffer, size_t size, HeapBlock* next) {

	_buffer = reinterpret_cast<uint8_t*>(buffer);
	_size = size;
	_next = next;
	_isExec = false;
	_lowestCodeBlock = NULL;

	NanoAssert((void*)this == (reinterpret_cast<const uint8_t*>(_buffer) + size - sizeof(HeapBlock)));

	createInternalCodeBlock();
}

void HeapBlock::createInternalCodeBlock()
{
	NanoAssert(getTypeAlign<NIns>() <= getTypeAlign<CodeBlock>());
	NanoAssert(getTypeAlign<NIns>() <= getTypeAlign<HeapBlock>());

	CodeBlock* codeBlock = reinterpret_cast<CodeBlock*>(_buffer);

	const size_t align			= getTypeAlign<NIns>();
	const size_t alignMask		= align - 1;
	size_t allocatedSize		= reinterpret_cast<uintptr_t>(this) - reinterpret_cast<uintptr_t>(_buffer);
	NanoAssert((_size - sizeof(HeapBlock)) == allocatedSize);

	size_t sizeofBlockHeader = CodeBlock::sizeOfBlockHeader();

	NanoAssert((sizeofBlockHeader & alignMask) == 0);
	size_t codeBlockSize = (allocatedSize - sizeofBlockHeader) & ~alignMask;

	uint8_t* startBuffer = reinterpret_cast<uint8_t*>(codeBlock) + sizeofBlockHeader;
	NIns* end = reinterpret_cast<NIns*>(startBuffer + codeBlockSize);
	codeBlock->initialize(this, end, NULL, NULL, NULL);

	_lowestCodeBlock = codeBlock;
}


}

