

#ifndef __nanojit_LirOptimizationPass__
#define __nanojit_LirOptimizationPass__

#include "nanojit.h"

namespace nanojit
{
class LIns;

template<class OptimizationPassFirst, class OptimizationPassSecond>
class ChainLirOptimizationPass
{
	OptimizationPassFirst*      first;
	OptimizationPassSecond*     second;

public:

	ChainLirOptimizationPass(OptimizationPassFirst* f, OptimizationPassSecond* s) 
	: first(f)
	, second(s) {
	}

	inline OptimizationPassFirst* getFirst() {
		return first;
	}

	inline OptimizationPassFirst* getSecond() {
		return second;
	}

	inline void begin() {
		first->begin();
		second->begin();
	}

	inline void end() {
		first->end();
		second->end();
	}

	inline void processLIns(LIns* src) {
		first->processLIns(src);
		second->processLIns(src);
	}
};

template<class OptimizationPassType>
void runOptimizationPass(OptimizationPassType* a_pOptimizationPass, LirChainWriter* a_pWriter)
{
	a_pOptimizationPass->begin();

	IntrusiveLinkedList::Iterator it = a_pWriter->getStorage()->getBegin();
	IntrusiveLinkedList::Iterator itEnd = a_pWriter->getStorage()->getEnd();

	while (it != itEnd)
	{
		LInsElement* pElement = it.getElement<LInsElement>();
		LIns* pIns = pElement->data();
		a_pOptimizationPass->processLIns(pIns);

		it.moveNext();
	}

	a_pOptimizationPass->end();
}

}

#endif // __nanojit_LirOptimizationPass__
