
#include "nanojit.h"

namespace nanojit
{

PeepholeMacher::PeepholeMacher() 
: opcodeTreeVector(NULL) {

}

PeepholeMacher::~PeepholeMacher() {
	shutdown();
}

void PeepholeMacher::initialize(Allocator* _allocator) {
	allocator = _allocator;

	NanoAssert(opcodeTreeVector == NULL);
	opcodeTreeVector = reinterpret_cast<TreeVector**>(allocator->alloc(sizeof(TreeVector*) * LIR_sentinel));
	memset(opcodeTreeVector, 0, sizeof(TreeVector*) * LIR_sentinel);
}

void PeepholeMacher::shutdown() {
	opcodeTreeVector = NULL;
	allocator = NULL;
}

void PeepholeMacher::addPersistantTree(const LOpcode* tree, size_t count, uint32_t value) {

	NanoUnusedVariable(count);
	LOpcode root = *tree;

	if (opcodeTreeVector[root] == NULL) {
		opcodeTreeVector[root] = new (allocator) TreeVector(allocator);
	}

	opcodeTreeVector[root]->push_back(Pair<const LOpcode*, uint32_t> (tree, value));
}


uint32_t PeepholeMacher::match(LIns* ins) {

	TreeVector* vector = opcodeTreeVector[ins->opcode()];

	if (vector != NULL) {

		for (TreeVector::Iter it = vector->begin(), itEnd = vector->end(); it != itEnd; ++it) {

			const LOpcode* result = matchTree(ins, it->first);

			if (result != NULL)
				return it->second;
		}
	}

	return MATCH_FAILED;
}


const LOpcode* PeepholeMacher::matchTree(LIns* ins, const LOpcode* tree) {

	const LOpcode* next = NULL;

	if (*tree == LIR_sentinel) // end of branch
		return tree + 1;

	if (ins->opcode() == *tree) {

		switch (ins->repKind()) {

			case LRK_Op1:
				if (ins->oprnd1()->canOptimize())
					next = matchTree(ins->oprnd1(), tree + 1);
				break;

			case LRK_Op2:
				if (ins->oprnd1()->canOptimize() && ins->oprnd2()->canOptimize()) {
					next = matchTree(ins->oprnd1(), tree + 1);
					if (next != NULL) {
						next = matchTree(ins->oprnd2(), next);
					}
				}
				break;

			default:
				break;
		}
	}


	return next;
}

}
